"""
Defines the core data models in peewee ORM
"""

import json
from datetime import date, timedelta, datetime
from xdg_base_dirs import xdg_data_home
from peewee import (
    BooleanField,
    CharField,
    DateTimeField,
    Model,
    SqliteDatabase,
    TextField,
)
from playhouse.sqlite_ext import JSONField

# Global DB connection
db = SqliteDatabase(
    xdg_data_home() / "tt/tt.db",
    pragmas={
        "journal_mode": "wal",
        "synchronous": "normal",
        "foreign_keys": 1,
    },
)


class BaseModel(Model):
    class Meta:
        database = db


class Thing(BaseModel):
    """
    Thing uses sqlite as a sort of schema-less database.

    - type corresponds to a table type in config.toml
    - data is the JSON data, which adheres (loosely) to the schema defined for the given type
    - also contains additional system-wide metadata fields tracked by app

    Note: There are a few places in the code where an attribute of Thing is
    accessed where the caller can't be sure if a field name is part of the
    model or somewhere within data.

    Therefore the lookup logic is almost always:

    - check for attribute first, e.g. getattr(obj, "field")
    - if the attribute does not exist, check for data["field"]
    - if neither is present, raise an error

    Therefore, `data` cannot contain properties with names used by attributes here.
    """

    type = CharField()
    data = JSONField()
    created_at = DateTimeField(default=datetime.now)
    updated_at = DateTimeField(default=datetime.now)
    deleted = BooleanField(default=False)

    def save(self, *args, **kwargs):
        # override of save method to ensure updated_at timestamp is always set
        self.updated_at = datetime.now()
        # schema enforcement could happen here as well, if/when needed
        return super(Thing, self).save(*args, **kwargs)

    @property
    def due_week(self):
        # FIXME: this is an experimental hack for spring quarter
        # ideally will have a way to define logic about this complex within config
        return self.due.isocalendar()[1] - 12


class SavedSearch(BaseModel):
    """
    Saved search parameters.

    Possibly to be removed in favor of config.toml saved searches?

    """

    name = CharField(unique=True)
    filters = CharField()
    sort_string = CharField()

    def __str__(self):
        return self.name


class ThingGenerator(BaseModel):
    """
    Abstract "thing generator" class, allows creation of recurring tasks
    and other "things that make other things".
    """

    template = CharField()
    type = CharField()
    config = TextField()  # TODO: JSON
    deleted = BooleanField(default=False)
    last_generated_at = DateTimeField(null=True)
    created_at = DateTimeField(default=datetime.now)

    def next_at(self) -> datetime.date:
        if self.deleted:
            return None

        today = date.today()
        val = int(json.loads(self.config)["val"])

        if self.type == "days-btwn":
            if not self.last_generated_at:
                return today
            return self.last_generated_at + timedelta(days=val)
        elif self.type == "monthly":
            year, month, day = today.year, today.month, today.day
            day_of_month = val

            # if we need to go forward a month
            if day_of_month < day:
                month += 1
                if month == 13:
                    month = 1
                    year += 1

            # recurring on 29-31 in Feb will just happen on 28th
            if day_of_month >= 29 and month == 2:
                maybe_next = date(year, month, 28)
            else:
                maybe_next = date(year, month, day_of_month)

            if not self.last_generated_at or self.last_generated_at < maybe_next:
                return maybe_next

            # FIXME: this doesn't handle if a month was missed somehow, just advances one
            # same logic as above, if we're stepping another month forward
            month += 1
            if month == 13:
                month = 1
                year += 1
            if day_of_month >= 29 and month == 2:
                maybe_next = date(year, month, 28)
            else:
                maybe_next = date(year, month, day_of_month)
            return maybe_next


def initialize_db():
    db.connect()
    db.create_tables([SavedSearch, ThingGenerator, Thing])
    db.close()
